# Copyright 2020 StackHut Limited (trading as Datapane)
# SPDX-License-Identifier: Apache-2.0
from .datafiles import ArrowFormat  # noqa: F401
from .utils import (  # noqa: F401
    guess_type,
    timestamp,
    unixtime,
    utf_read_text,
)



from __future__ import annotations

import dataclasses
import sys
from collections import namedtuple
from typing import TYPE_CHECKING, Any, Protocol, cast

from multimethod import DispatchError, multimethod

from arakawa import schemas
from arakawa.blocks import BaseBlock, Group
from arakawa.blocks.asset import AssetBlock
from arakawa.blocks.layout import ContainerBlock
from arakawa.blocks.text import EmbeddedTextBlock
from arakawa.exceptions import ARError
from arakawa.types import VAlign
from arakawa.utils import log
from arakawa.view.view_blocks import Blocks
from arakawa.view.visitors import ViewVisitor

if sys.version_info <= (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

if TYPE_CHECKING:
    from arakawa.processors import FileEntry, FileStore


@dataclasses.dataclass
class PydanticBuilder(ViewVisitor):
    store: FileStore
    elements: list[BaseBlock] = dataclasses.field(default_factory=list)

    _seen_names: set[str] = dataclasses.field(default_factory=set)

    def get_root(self, fragment: bool = False):
        _top_group = cast(Group, self.elements.pop())
        assert _top_group._type == "Group"
        assert not self.elements

        return schemas.View.model_validate(
            {
                "_type": "View",
                "fragment": fragment,
                "version": 1,
                "blocks": _top_group.blocks,
            }
        )

    @property
    def store_count(self) -> int:
        return len(self.store.files)

    def add_element(self, _: BaseBlock, e: BaseBlock) -> Self:
        """Add an element to the list of nodes at the current XML tree location"""
        self.elements.append(e)

        name: str | None = getattr(e, "name", None)
        if name:
            if name in self._seen_names:
                raise ARError(f"Duplicate name {name} found in the View")

            self._seen_names.add(name)

        return self

    @multimethod
    def visit(self, b: BaseBlock) -> Self:
        """Base implementation - just created an empty tag including all the initial attributes"""
        return self.add_element(b, b)

    def _visit_subnodes(self, b: ContainerBlock):
        cur_elements = self.elements
        self.elements = []
        b.traverse(self)  # visit subnodes
        res = self.elements
        self.elements = cur_elements
        return res

    @visit.register  # type: ignore
    def _(self, b: ContainerBlock) -> Self:
        sub_elements = self._visit_subnodes(b)
        b._add_attributes(blocks=sub_elements)
        return self.add_element(b, b)

    @visit.register  # type: ignore
    def _(self, b: Blocks) -> Self:
        sub_elements = self._visit_subnodes(b)

        # Blocks are converted to Group internally
        if label := getattr(b, "label", None):
            log.info(f"Found label {label} in top-level Blocks/View")

        element = Group(
            blocks=sub_elements,
            name=b.name,
            label=label,
            columns=1,
            valign=VAlign.TOP,
        )
        return self.add_element(b, element)

    @visit.register  # type: ignore
    def _(self, b: EmbeddedTextBlock) -> Self:
        return self.add_element(b, b)

    @visit.register  # type: ignore
    def _(self, b: AssetBlock):
        fe = self._add_asset_to_store(b)

        element = b.copy()
        element._add_attributes(
            type=fe.mime,
            src=f"ref://{fe.hash}",
        )
        return self.add_element(b, element)

    def _add_asset_to_store(self, b: AssetBlock) -> FileEntry:
        """Default asset store handler that operates on native Python objects"""
        # import here as a very slow module due to nested imports
        # from .. import files

        # check if we already have stored this asset to the store
        # TODO - do we just persist the asset store across the session??
        if b._prev_entry:
            if type(b._prev_entry) == self.store.fw_klass:  # noqa: E721
                self.store.add_file(b._prev_entry)
                return b._prev_entry
            b._prev_entry = None

        if b.data is not None:
            # fe = files.add_to_store(self.data, s.store)
            try:
                writer = get_writer(b)
                meta: AssetMeta = writer.get_meta(b.data)
                fe = self.store.get_file(meta.ext, meta.mime)
                writer.write_file(b.data, fe.file)
                self.store.add_file(fe)
            except DispatchError as e:
                raise ARError(
                    f"{type(b.data).__name__} not supported for {self.__class__.__name__}"
                ) from e
        elif b.file is not None:
            fe = self.store.load_file(b.file)
        else:
            raise ARError("No asset to add")

        b._prev_entry = fe
        return fe


AssetMeta = namedtuple("AssetMeta", "ext mime")


class AssetWriterP(Protocol):
    """Implement these in any class to support asset writing
    for a particular AssetBlock"""

    def get_meta(self, x: Any) -> AssetMeta: ...

    def write_file(self, x: Any, f) -> None: ...


def get_writer(b: AssetBlock) -> AssetWriterP:
    import arakawa.blocks.asset as a

    from . import asset_writers as aw

    if isinstance(b, a.Plot):
        return aw.PlotWriter()

    if isinstance(b, a.Table):
        return aw.HTMLTableWriter()  # type: ignore

    if isinstance(b, a.Attachment):
        return aw.AttachmentWriter()

    if isinstance(b, a.DataTable):
        return aw.DataTableWriter()  # type: ignore

    raise KeyError(f"No writer found for {type(b).__name__}")



from .asset import Attachment, DataTable, Media, Plot, Table  # noqa: F401
from .base import DataBlock  # noqa: F401
from .controls import (  # noqa: F401
    ChoiceField,
    DateTimeField,
    FileField,
    HiddenField,
    MultiChoiceField,
    NumberBox,
    RangeField,
    SwitchField,
    TagsField,
    TextBox,
)
from .layout import Compute, Group, Page, Select, Toggle, View  # noqa: F401
from .misc_blocks import BigNumber, Empty  # noqa: F401
from .text import Code, Embed, Text  # noqa: F401



from __future__ import annotations

from typing import Literal

from pydantic import AnyUrl, Field

from .base import DataBlock
from .fields import TypeAliasedField
from .mixins import OptionalCaptionMixin, OptionalLabelMixin, OptionalNameMinx


class AssetBlock(OptionalLabelMixin, OptionalNameMinx, OptionalCaptionMixin, DataBlock):
    src: AnyUrl
    type: str | None = Field(default=None)  # pattern=r"\w+/[\w.+\-]+"


class Media(AssetBlock):
    type_: Literal["Media"] = TypeAliasedField()


class Attachment(AssetBlock):
    type_: Literal["Attachment"] = TypeAliasedField()

    filename: str = Field(..., min_length=1, max_length=128)


class Plot(AssetBlock):
    type_: Literal["Plot"] = TypeAliasedField()

    scale: float = Field(default=1.0, ge=0.0)
    responsive: bool = Field(default=True)


class Table(AssetBlock):
    type_: Literal["Table"] = TypeAliasedField()


class DataTable(AssetBlock):
    type_: Literal["DataTable"] = TypeAliasedField()

    rows: int = Field(..., ge=0)
    columns: int = Field(..., ge=0)



from __future__ import annotations

from functools import partial
from typing import Annotated

from pydantic import BeforeValidator, Field

from .base import DataBlock


def truncate_optional_string(x: str | None, *, max_length: int) -> str | None:
    if x is None or len(x) <= max_length:
        return x

    return f"{x[: max_length - 3]}..."


class OptionalLabelMixin(DataBlock):
    label: Annotated[
        str | None, BeforeValidator(partial(truncate_optional_string, max_length=256))
    ] = Field(
        default=None,
        description="A label to display above the block",
        min_length=1,
        max_length=256,
    )


class OptionalCaptionMixin(DataBlock):
    caption: Annotated[
        str | None, BeforeValidator(partial(truncate_optional_string, max_length=512))
    ] = Field(
        default=None,
        description="A caption to display below the block",
        min_length=1,
        max_length=512,
    )


class OptionalNameMinx(DataBlock):
    name: str | None = Field(
        default=None,
        description="A unique name for the block to reference when adding text or embedding",
    )


class NameMixin(DataBlock):
    name: str = Field(
        ...,
        description="A unique name for the block to reference when adding text or embedding",
    )



"""
Conversion templates for IPython notebooks to Arakawa apps.

..note: Early stage feature - a diverse set of templates that are subject to change.
  Most implementations are currently low-level and will be replaced with higher-level abstractions over time.
"""

from __future__ import annotations

from abc import abstractmethod
from typing import Callable, Union

import arakawa.blocks as b
from arakawa.ipython.exceptions import BlocksNotFoundError
from arakawa.utils import display_msg

BlockFilterF = Callable[[b.BaseBlock], bool]
BlockTypes = Union[tuple[type[b.BaseBlock], ...], type]
BaseElementList = list[b.BaseBlock]

_registry: dict[str, type[IPythonTemplate]] = {}


def partition_blocks_by_predicates(
    blocks: BaseElementList, predicates: list[BlockFilterF]
) -> list[BaseElementList]:
    """Partition blocks by predicates"""
    partitions: list[BaseElementList] = [[] for _ in range(len(predicates))]

    for block in blocks:
        for i, pred in enumerate(predicates):
            if pred(block):
                partitions[i].append(block)
                break

    return partitions


def partition_blocks_by_types(
    blocks: BaseElementList, partition_types: list[type[b.BaseBlock]]
) -> list[BaseElementList]:
    """Partition blocks by types"""

    predicates: list[BlockFilterF] = [
        lambda block, partition_type=partition_type: isinstance(block, partition_type)  # type: ignore # https://github.com/python/mypy/issues/12557, https://github.com/python/mypy/issues/4226
        for partition_type in partition_types
    ]
    partitions: list[BaseElementList] = partition_blocks_by_predicates(
        blocks, predicates
    )

    return partitions


def filter_blocks_by_predicate(
    blocks: BaseElementList, predicate: BlockFilterF
) -> BaseElementList:
    """Filter blocks by predicates"""
    filtered_blocks, *_ = partition_blocks_by_predicates(blocks, [predicate])

    return filtered_blocks


def filter_blocks_by_types(
    blocks: BaseElementList, block_types: BlockTypes
) -> BaseElementList:
    """Filter blocks by type"""
    filtered_blocks, *_ = partition_blocks_by_predicates(
        blocks, [lambda block: isinstance(block, block_types)]
    )

    return filtered_blocks


def guess_template(blocks: BaseElementList) -> type[IPythonTemplate]:
    """Guess the template to use based on the blocks provided"""
    app_template: type[IPythonTemplate]

    # DashboardTemplate: Contains only Plot, BigNumber, and DataTable blocks
    if all(isinstance(block, (b.Plot, b.BigNumber, b.DataTable)) for block in blocks):
        app_template = DashboardTemplate
    # TitledPagesTemplate: Contains text blocks with at least two headings
    elif (
        len(
            filter_blocks_by_predicate(
                blocks,
                lambda block: isinstance(block, b.Text)
                and block.content.startswith("# "),
            )
        )
        >= 2
    ):
        app_template = TitledPagesTemplate
    # DescriptivePagesTemplate: Contains at least two text blocks that are followed by a different block type
    elif (
        "".join(
            ["1" if isinstance(block, (b.Text)) else "0" for block in blocks]
        ).count("10")
        >= 2
    ):
        app_template = DescriptivePagesTemplate
    # AssetListTemplate: Does not contain any text or code blocks
    elif not any(filter_blocks_by_types(blocks, (b.Text, b.Code))):
        app_template = AssetListTemplate
    # AssetCodeListTemplate: Does not contain any text blocks
    elif not any(filter_blocks_by_types(blocks, b.Text)):
        app_template = AssetCodeListTemplate
    # ReportTemplate: The default template if we can't make a guess
    else:
        app_template = ReportTemplate

    display_msg(
        f"Automatically selecting the `{app_template.name}` template. You can override this with `template=template_name` from the following templates: {', '.join(_registry.keys())}."
    )

    return app_template


class IPythonTemplate:
    name: str = "IPythonTemplate"

    def __init__(self, blocks):
        self.blocks = blocks

    @classmethod
    def __init_subclass__(cls, template_name, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.name = template_name
        _registry[template_name] = cls

    @abstractmethod
    def transform(self) -> None:
        raise NotImplementedError

    def validate(self):
        if not self.blocks:
            raise BlocksNotFoundError("No blocks required by template were found.")


class ReportTemplate(IPythonTemplate, template_name="report"):
    """Creates a report with blocks in the given order"""

    def transform(self) -> None:
        pass


class DashboardTemplate(IPythonTemplate, template_name="dashboard"):
    """Creates a dashboard with a 3 column group of BigNumbers, followed by a 2 column group of Plots"""

    def transform(self) -> None:
        blocks: BaseElementList = self.blocks

        big_numbers, plots, tables = partition_blocks_by_types(
            blocks,
            [b.BigNumber, b.Plot, b.DataTable],
        )

        blocks = []

        if big_numbers:
            blocks.append(b.Group(blocks=big_numbers, columns=3))

        tabs: BaseElementList = []

        if plots:
            tabs.append(b.Group(blocks=plots, columns=2, label="Figures"))

        if tables:
            tabs.append(b.Group(blocks=tables, label="Tables"))

        # We may only have plots or tables, so we don't need to use a Select block
        if tabs:
            if len(tabs) > 1:
                blocks.append(b.Select(blocks=tabs))
            else:
                blocks.append(tabs[0])

        self.blocks = blocks


class AssetListTemplate(IPythonTemplate, template_name="asset_list"):
    """Starts a new page for every supported Arakawa block"""

    def transform(self) -> None:
        blocks = filter_blocks_by_predicate(
            self.blocks, lambda block: not isinstance(block, (b.Code, b.Text))
        )
        pages: list[b.BaseBlock] = [
            b.Page(blocks=[block], title=f"{idx + 1}. {block._tag}")
            for idx, block in enumerate(blocks)
        ]
        self.blocks = pages


class AssetCodeListTemplate(IPythonTemplate, template_name="asset_code_list"):
    """Starts a new page for every supported Arakawa block, with the code block after the asset"""

    def transform(self) -> None:
        blocks = self.blocks
        pages: list[b.BaseBlock] = []
        last_block: b.BaseBlock | None = None

        for block in blocks:
            # If the block is not a Code or Text block, add it to a new page
            if not isinstance(block, (b.Text, b.Code)):
                # If the last block was a code block, add it to the current page
                if isinstance(last_block, b.Code):
                    pages.append(
                        b.Page(
                            blocks=[b.Group(block, last_block)],
                            title=f"{len(pages) + 1}. {block._tag}",
                        )
                    )
                else:
                    pages.append(
                        b.Page(blocks=[block], title=f"{len(pages) + 1}. {block._tag}")
                    )

            last_block = block
        self.blocks = pages


class DescriptivePagesTemplate(IPythonTemplate, template_name="descriptive_pages"):
    """Start a new page every time a contiguous block of text is encountered"""

    def transform(self) -> None:
        blocks = self.blocks
        pages: list[b.BaseBlock] = []
        page_blocks: BaseElementList = []
        last_block: b.BaseBlock | None = None
        page_title: str = "Page 1"

        for block in blocks:
            # If the block is a Text block, and the last block was not a Text block, start a new page
            if isinstance(block, b.Text) and not isinstance(last_block, b.Text):
                if page_blocks:
                    pages.append(b.Page(blocks=page_blocks, title=page_title))
                    page_blocks = []

                if block.content.startswith("# "):
                    page_title = block.content.partition("\n")[0].lstrip("# ").strip()
                else:
                    page_title = f"Page {len(pages) + 1}"

            page_blocks.append(block)
            last_block = block

        # add the last page
        pages.append(b.Page(blocks=page_blocks, title=page_title))
        self.blocks = pages


class TitledPagesTemplate(IPythonTemplate, template_name="titled_pages"):
    """Start a new page every time heading (#) text is encountered"""

    def transform(self) -> None:
        blocks = self.blocks
        pages: list[b.BaseBlock] = []
        page_blocks: BaseElementList = []
        page_title: str = "Page 1"

        for block in blocks:
            # If the block is a Text block, and it starts with a # heading, start a new page
            if isinstance(block, b.Text) and block.content.startswith("# "):
                if page_blocks:
                    pages.append(b.Page(blocks=page_blocks, title=page_title))
                    page_blocks = []

                page_title = block.content.partition("\n")[0].lstrip("#").strip()

            page_blocks.append(block)

        # add the last page
        pages.append(b.Page(blocks=page_blocks, title=page_title))
        self.blocks = pages



from .pydantic_visitor import PydanticBuilder  # noqa: F401
from .view_blocks import Blocks, BlocksT, Report, View  # noqa: F401
from .visitors import PreProcess, PrettyPrinter, ViewVisitor  # noqa: F401



from .api import (  # noqa: F401
    save_report,
    stringify_report,
)
from .file_store import FileEntry, FileStore  # noqa: F401
from .processors import ConvertPydantic, PreProcessView  # noqa: F401
from .types import (  # noqa: F401
    FontChoice,
    Formatting,
    Pipeline,
    TextAlignment,
    ViewState,
    Width,
    mk_null_pipe,
)



from __future__ import annotations

import abc
import datetime
import gzip
import hashlib
import io
import sys
import tempfile
from pathlib import Path
from shutil import copyfileobj
from typing import IO, BinaryIO

import base64io

if sys.version_info <= (3, 11):
    from typing_extensions import Self
else:
    from typing import Self


from arakawa.common import guess_type

SERVED_REPORT_ASSETS_DIR = "assets"
GZIP_MTIME = datetime.datetime(year=2000, month=1, day=1).timestamp()


class FileEntry:
    file: IO
    _ext: str
    _dir_path: Path | None

    # post-freeze
    frozen: bool = False
    mime: str
    hash: str
    size: int
    wrapped: BinaryIO

    def __init__(self, ext: str, mime: str | None = None, dir_path: Path | None = None):
        self.mime = mime or guess_type(Path(f"tmp{ext}"))
        self._ext = ext
        self._dir_path = dir_path

    @abc.abstractmethod
    def freeze(self) -> None:
        """Must be called after writing / adding to store
        # TODO - add to contextmanager??
        """

    @property
    @abc.abstractmethod
    def src(self) -> str:
        pass

    def as_dict(self) -> dict:
        assert self.frozen
        return {
            "src": self.src,
            "hash": self.hash,
            "size": self.size,
            "mime": self.mime,
        }

    def __eq__(self, other: FileEntry) -> bool:
        if self.hash:
            return self.hash == other.hash

        raise NotImplementedError()


class NullWriter(io.BytesIO):
    def write(self, s):
        pass

    def writelines(self, *a, **kw) -> None:
        pass


class DummyFileEntry(FileEntry):
    """File entry that discards all data - for internal use"""

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self.file = NullWriter()

    def freeze(self):
        self.size = 0
        self.hash = "abcdef"
        self.frozen = True

    def src(self) -> str:
        return "/dev/null"


class B64FileEntry(FileEntry):
    """Memory-based b64 file"""

    # requires b64io is bytes only and wraps to a bytes file only
    file: base64io.Base64IO
    wrapped: io.BytesIO
    contents: bytes

    def __init__(self, ext: str, mime: str | None = None, *a, **kw):
        super().__init__(ext, mime, *a, **kw)
        self.wrapped = io.BytesIO()
        self.file = base64io.Base64IO(self.wrapped)

    def freeze(self) -> None:
        if not self.frozen:
            self.frozen = True
            # get a reference to the buffer to splice later
            self.file.close()
            self.file.flush()
            self.contents = self.wrapped.getvalue()
            # calc other properties
            self.hash = hashlib.sha256(self.contents).hexdigest()[:10]
            self.size = self.wrapped.tell()

    @property
    def src(self) -> str:
        return f"data:{self.mime};base64,{self.contents.decode('ascii')}"


class GzipTmpFileEntry(FileEntry):
    """Gzipped file, by default stored in /tmp"""

    # both file and wapper files are bytes-only
    file: gzip.GzipFile
    # TODO - this could actually be an in-memory file...
    wrapped: tempfile._TemporaryFileWrapper
    has_output_dir: bool = False

    # Do we need DPTmpFile here, or just use namedtempfile??
    def __init__(self, ext: str, mime: str | None = None, dir_path: Path | None = None):
        super().__init__(ext, mime, dir_path)

        if dir_path:
            # create as a permanent file within the given dir
            self.has_output_dir = True
            self.wrapped = tempfile.NamedTemporaryFile(  # noqa: SIM115
                "w+b", suffix=ext, prefix="ar-", dir=dir_path, delete=False
            )
        else:
            self.wrapped = tempfile.NamedTemporaryFile("w+b", suffix=ext, prefix="ar-")  # noqa: SIM115

        self.file = gzip.GzipFile(fileobj=self.wrapped, mode="w+b", mtime=GZIP_MTIME)

    def calc_hash(self, f: IO) -> str:
        f.seek(0)
        file_hash = hashlib.sha256()
        while chunk := f.read(8192):
            file_hash.update(chunk)
        return file_hash.hexdigest()[:10]

    @property
    def src(self) -> str:
        if self.has_output_dir:
            return f"/{SERVED_REPORT_ASSETS_DIR}/{Path(self.wrapped.name).name}"
        return "NYI"

    def freeze(self) -> None:
        if not self.frozen:
            self.frozen = True
            self.file.flush()
            self.file.close()
            self.wrapped.flush()
            # size will be the compressed size...
            self.size = self.wrapped.tell()
            self.hash = self.calc_hash(self.wrapped)


class FileStore:
    # TODO - make this a CAS (index by object hash itself?)
    # NOTE - currently we pass dir_path via the FileStore, could move into the file themselves?
    def __init__(self, fw_klass: type[FileEntry], assets_dir: Path | None = None):
        super().__init__()
        self.fw_klass = fw_klass
        self.files: list[FileEntry] = []
        self.dir_path = assets_dir

    def __add__(self, other: Self) -> Self:
        # TODO - ensure factory is the same for both
        self.files.extend(other.files)
        return self

    @property
    def store_count(self) -> int:
        return len(self.files)

    @property
    def file_list(self) -> list[BinaryIO]:
        return [f.wrapped for f in self.files]

    def get_file(self, ext: str, mime: str) -> FileEntry:
        return self.fw_klass(ext, mime, self.dir_path)

    def add_file(self, fw: FileEntry) -> None:
        fw.freeze()
        self.files.append(fw)

    def load_file(self, path: Path) -> FileEntry:
        """load a file into the store (makes a copy)"""
        # TODO - ideally lazily-link a path to the store (rather than copy it in)
        ext = "".join(path.suffixes)
        dest_obj = self.fw_klass(ext=ext, dir_path=self.dir_path)
        with path.open("rb") as src_obj:
            copyfileobj(src_obj, dest_obj.file)
        self.add_file(dest_obj)
        return dest_obj

    def as_dict(self) -> dict:
        """Build a json structure suitable for embedding in a html file, json-rpc response, etc."""
        x: FileEntry  # noqa: F842
        return {x.hash: x.as_dict() for x in self.files}

    def get_entry(self, hash: str) -> FileEntry | None:
        # TODO - change self.files to a dict[hash, FileEntry]?
        return next((f for f in self.files if f.hash == hash), None)



# Copyright 2020 StackHut Limited (trading as Datapane)
# SPDX-License-Identifier: Apache-2.0
import importlib.metadata

__version__ = importlib.metadata.version(__name__)

# Public API re-exports
from .blocks import (  # noqa: F401
    HTML,
    Attachment,
    BigNumber,
    Block,
    ChoiceField,
    Code,
    ColorField,
    Compute,
    DataTable,
    DateField,
    DateTimeField,
    EmailField,
    Embed,
    Empty,
    FileField,
    Formula,
    Group,
    HiddenField,
    Media,
    MultiChoiceField,
    NumberBox,
    Page,
    PasswordField,
    Plot,
    RangeField,
    SearchField,
    Select,
    SelectType,
    SwitchField,
    Table,
    TagsField,
    TelephoneField,
    Text,
    TextareaField,
    TextBox,
    TimeField,
    Toggle,
    URLField,
    VAlign,
    wrap_block,
)
from .exceptions import ARError  # noqa: F401
from .processors import (  # noqa: F401
    FontChoice,
    Formatting,
    TextAlignment,
    Width,
    save_report,
    stringify_report,
)
from .utils import enable_logging, print_debug_info  # noqa: F401
from .view import Blocks, Report, View  # noqa: F401

from . import builtins  # noqa: F401 # isort: skip

X = wrap_block



from arakawa.utils import display_msg


class NotebookError(Exception):
    """Exception raised when a Notebook to Arakawa conversion fails."""

    def _render_traceback_(self):
        display_msg(
            f"""**Conversion failed**

{self!s}"""
        )


class NotebookParityError(NotebookError):
    """Exception raised when IPython output cache is not in sync with the saved notebook"""


class BlocksNotFoundError(NotebookError):
    """Exception raised when no blocks are found during conversion"""



from __future__ import annotations

import sys
from collections import deque
from functools import reduce
from typing import TYPE_CHECKING

from arakawa.exceptions import ARError
from arakawa.types import ComputeMethod, SelectType, VAlign
from arakawa.utils import log

from .base import BaseBlock, BlockId, BlockList, BlockOrPrimitive, wrap_block
from .misc_blocks import Empty, gen_name

if sys.version_info <= (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

if TYPE_CHECKING:
    from arakawa.blocks import Block

    from .base import VV


class ContainerBlock(BaseBlock):
    """
    Abstract Block that supports nested/contained blocks
     - represents a subtree in the document
    """

    blocks: BlockList
    # how many blocks must there be in the container
    report_minimum_blocks: int = 1

    def __init__(
        self,
        *arg_blocks: BlockOrPrimitive,
        blocks: list[BlockOrPrimitive] | None = None,
        **kwargs,
    ):
        self.blocks = [wrap_block(b) for b in blocks or list(arg_blocks)]

        super().__init__(**kwargs)

    def __iter__(self):
        return BlockListIterator(self.blocks.__iter__())

    def __add__(self, other: Self) -> Self:
        self.blocks.extend(other.blocks)
        return self

    def __and__(self, other: Self) -> Self:
        self.blocks.extend(other.blocks)
        return self

    def __copy__(self) -> Self:
        inst = super().copy()
        inst.blocks = self.blocks.copy()
        return inst

    @classmethod
    def empty(cls) -> Self:
        return cls(blocks=[Empty(gen_name())])

    def traverse(self, visitor: VV) -> VV:
        # perform a depth-first traversal of the contained blocks
        return reduce(
            lambda _visitor, block: block.accept(_visitor), self.blocks, visitor
        )


class Page(ContainerBlock):
    """
    Apps on Arakawa can have multiple pages, which are presented to users as tabs at the top of your app. These can be used similarly to sheets in an Excel document.

    To add a page, use the `ar.Page` block at the top-level of your app, and give it a title with the `title` parameter.

    !!! info
        Pages cannot be nested, and can only exist at the root level of your `ar.App` object. If you're using pages, all other blocks must be contained inside a Page block.

    !!! note
        This is included for backwards-compatibility, and can be replaced by using Selects going forwards.
    """

    # BC-only helper - converted into a Select + Group within the post-XML processor
    _tag = "_Page"

    def __init__(
        self,
        *arg_blocks: BlockOrPrimitive,
        blocks: list[BlockOrPrimitive] | None = None,
        title: str | None = None,
        name: BlockId | None = None,
    ):
        """
        Args:
            *arg_blocks: Blocks to add to Page
            blocks: Allows providing the report blocks as a single list
            title: The page title (optional)
            name: A unique id for the Page to aid querying (optional)

        !!! tip
            Page can be passed using either arg parameters or the `blocks` kwarg, e.g. `ar.Page(group, select)` or `ar.Group(blocks=[group, select])`
        """
        self.title = title
        super().__init__(*arg_blocks, blocks=blocks, label=title, name=name)

        if any(isinstance(b, Page) for b in self.blocks):
            raise ARError("Nested pages not supported, please use Selects and Groups")


class Select(ContainerBlock):
    """
    Selects act as a container that holds a list of nested Blocks objects, such
    as Tables, Plots, etc.. - but only one may be __visible__, or "selected", at once.

    The user can choose which nested object to view dynamically using either tabs or a dropdown.

    !!! note
        Select expects a list of Blocks, e.g. a Plot or Table, but also includes Select or Groups themselves, but if a Python object is passed, e.g. a Dataframe, Arakawa will attempt to convert it automatically.

    """

    _tag = "Select"
    report_minimum_blocks = 2

    def __init__(
        self,
        *arg_blocks: BlockOrPrimitive,
        blocks: list[BlockOrPrimitive] | None = None,
        type: SelectType = SelectType.TABS,
        name: BlockId | None = None,
        label: str | None = None,
    ):
        """
        Args:
            *arg_blocks: Page to add to report
            blocks: Allows providing the report blocks as a single list
            type: An instance of SelectType that indicates if the select should use tabs or a dropdown (default: Tabs)
            name: A unique id for the blocks to aid querying (optional)
            label: A label used when displaying the block (optional)

        !!! tip
            Select can be passed using either arg parameters or the `blocks` kwarg, e.g. `ar.Select(table, plot, type=ar.SelectType.TABS)` or `ar.Select(blocks=[table, plot])`
        """
        super().__init__(*arg_blocks, blocks=blocks, name=name, label=label, type=type)
        if len(self.blocks) < 2:
            log.info("Creating a Select with less than 2 objects")


class Group(ContainerBlock):
    """
    If you pass a list of blocks (such as `Plot` and `Table`) to an app, they are -- by default -- laid out in a single column with a row per block.

    If you would like to customize the rows and columns, Arakawa provides a `Group` block which takes a list of blocks and a number of columns and lays them out in a grid.

    !!! tip
        As `Group` blocks are blocks themselves, they are composable, and you can create more custom layers of nested blocks, for instance nesting 2 rows in the left column of a 2 column layout
    """

    _tag = "Group"

    def __init__(
        self,
        *arg_blocks: BlockOrPrimitive,
        blocks: list[BlockOrPrimitive] | None = None,
        name: BlockId | None = None,
        label: str | None = None,
        widths: list[int | float] | None = None,
        valign: VAlign = VAlign.TOP,
        columns: int = 1,
    ):
        """
        Args:
            *arg_blocks: Group to add to report
            blocks: Allows providing the report blocks as a single list
            name: A unique id for the blocks to aid querying (optional)
            label: A label used when displaying the block (optional)
            widths: A list of numbers representing the proportion of vertical space given to each column (optional)
            valign: The vertical alignment of blocks in the Group (default = VAlign.TOP)
            columns: Display the contained blocks, e.g. Plots, using _n_ columns (default = 1), setting to 0 auto-wraps the columns

        !!! note
            Group can be passed using either arg parameters or the `blocks` kwarg, e.g. `ar.Group(plot, table, columns=2)` or `ar.Group(blocks=[plot, table], columns=2)`.
        """

        if widths is not None and len(widths) != columns:
            raise ARError("Group 'widths' list length does not match number of columns")

        # columns = columns or len(self.blocks)
        self.columns = columns
        super().__init__(
            *arg_blocks,
            blocks=blocks,
            name=name,
            label=label,
            widths=widths,
            valign=valign,
        )


class Toggle(ContainerBlock):
    """
    Toggles act as a container that holds a list of nested Block objects, whose visibility can be toggled on or off by the report viewer
    """

    _tag = "Toggle"
    report_minimum_blocks = 1

    def __init__(
        self,
        *arg_blocks: BlockOrPrimitive,
        blocks: list[BlockOrPrimitive] | None = None,
        name: BlockId | None = None,
        label: str | None = None,
    ):
        """
        Args:
            *arg_blocks: Group to add to report
            blocks: Allows providing the report blocks as a single list
            name: A unique id for the blocks to aid querying (optional)
            label: A label used when displaying the block (optional)
        """
        super().__init__(*arg_blocks, blocks=blocks, name=name, label=label)
        self._wrap_blocks()

    def _wrap_blocks(self) -> None:
        """Wrap the list of blocks in a top-level block element if needed"""
        if len(self.blocks) > 1:
            # only wrap if not all blocks are a Group object
            self.blocks = [Group(blocks=self.blocks)]


class Compute(ContainerBlock):
    """
    Compute acts as a container that holds a list of ControlBlock objects to compose an HTML form.
    """

    _tag = "Compute"

    def __init__(
        self,
        *arg_blocks: BlockOrPrimitive,
        blocks: list[BlockOrPrimitive] | None = None,
        name: BlockId | None = None,
        label: str | None = None,
        prompt: str | None = None,
        subtitle: str | None = None,
        action: str | None = "",
        method: ComputeMethod = ComputeMethod.GET,
    ):
        """
        Args:
            *arg_blocks: Compute to add to report.
            blocks (list[BlockOrPrimitive] | None, optional): Blocks to compose a form. Defaults to None.
            name (BlockId | None, optional): A name. Defaults to None.
            label (str | None, optional): A label. Defaults to None.
            prompt (str | None, optional): A prompt. Defaults to None.
            subtitle (str | None, optional): A subtitle. Defaults to None.
            action (str | None, optional): The form action. Defaults to "".
            method (MethodType, optional): The form method. Defaults to MethodType.GET.
        """
        super().__init__(
            *arg_blocks,
            action=action or "",
            method=method,
            blocks=blocks,
            name=name,
            label=label,
            prompt=prompt,
            subtitle=subtitle,
        )


class BlockListIterator:
    """
    Wrapper around default list iterator that supports depth-first traversal of blocks.
    """

    def __init__(self, _iter):
        # linearize all blocks into a deque as we traverse
        self.nested = deque(_iter)

    def __next__(self) -> Block:
        try:
            b: Block = self.nested.popleft()
        except IndexError as e:
            raise StopIteration() from e

        if isinstance(b, ContainerBlock):
            # add the nested iter contents for next "next" call, hence traversing the tree
            self.nested.extendleft(reversed(b.blocks))
        return b

    def __iter__(self):
        return self



from __future__ import annotations

import dataclasses as dc
import datetime
import importlib.resources as ir
import locale
import math
import mimetypes
import re
import sys
import time
from collections.abc import Sized
from numbers import Number
from pathlib import Path
from typing import Any

import chardet
from chardet.universaldetector import UniversalDetector
from loguru import logger as log
from micawber import ProviderException, bootstrap_basic, bootstrap_noembed, cache

from arakawa.exceptions import ARError
from arakawa.types import HTML, MIME

################################################################################
# CONSTANTS
ON_WINDOWS = sys.platform == "win32"

################################################################################
# MIME-type handling
mimetypes.init(files=[str(ir.files("arakawa.resources") / "mime.types")])

# TODO - hardcode as temporary fix until mimetypes double extension issue is sorted
_double_ext_map = {
    ".vl.json": "application/vnd.vegalite.v5+json",
    ".vl5.json": "application/vnd.vegalite.v5+json",
    ".bokeh.json": "application/vnd.bokeh.show+json",
    ".pl.json": "application/vnd.plotly.v1+json",
    ".fl.html": "application/vnd.folium+html",
    ".tbl.html": "application/vnd.arakawa.table+html",
    ".tar.gz": "application/x-tgz",
}
double_ext_map: dict[str, MIME] = {k: MIME(v) for k, v in _double_ext_map.items()}


def guess_type(filename: Path) -> MIME:
    ext = "".join(filename.suffixes)
    if ext in double_ext_map:
        return double_ext_map[ext]
    mtype, _ = mimetypes.guess_type(str(filename))
    assert mtype
    return MIME(mtype or "application/octet-stream")


def guess_encoding(fn: str) -> str:
    with open(fn, "rb") as f:
        detector = UniversalDetector()
        for line in f.readlines():
            detector.feed(line)
            if detector.done:
                break
        detector.close()
    return detector.result["encoding"]  # type: ignore


def utf_read_text(file: Path) -> str:
    """Encoding-aware text reader
    - handles cases like on Windows where a file is UTF-8, but default locale is windows-1252
    """
    if ON_WINDOWS:
        f_bytes = file.read_bytes()
        f_enc: str = chardet.detect(f_bytes)["encoding"]  # type: ignore
        # NOTE - can just special case utf-8 files here?
        def_enc = locale.getpreferredencoding()
        log.debug(f"Default encoding is {def_enc}, file encoded as {f_enc}")
        if def_enc.upper() != f_enc.upper():
            log.warning(f"Text file {file} encoded as {f_enc}, auto-converting")
        return f_bytes.decode(encoding=f_enc)
    # for linux/macOS assume utf-8
    return file.read_text()


def unixtime() -> int:
    return int(time.time())


def timestamp(x: datetime.datetime | None = None) -> str:
    """Return ISO timestamp for a datetime"""
    x = x or datetime.datetime.utcnow()
    return f"{x.isoformat(timespec='seconds')}{'' if x.tzinfo else 'Z'}"


def is_valid_id(id: str) -> bool:
    """(cached) regex to check for a xsd:ID name"""
    return re.fullmatch(r"^[a-zA-Z_][\w.-]*$", id) is not None


def conv_attrib(v: Any) -> Any | None:
    """
    Convert a value to a str for use as an ElementBuilder attribute
    - also handles None to a string for optional field values
    """
    # TODO - use a proper serialization framework here / lxml features
    if v is None:
        return v

    if isinstance(v, Sized) and len(v) == 0:
        return None

    if isinstance(v, str):
        return v

    if isinstance(v, Number) and not isinstance(v, bool):
        if math.isinf(v) and v > 0:  # type: ignore
            return "INF"

        if math.isinf(v) and v < 0:  # type: ignore
            return "-INF"

        if math.isnan(v):  # type: ignore
            return "NaN"

        return str(v)

    return v


def mk_attribs(**attribs: Any):
    """convert attributes, dropping None and empty values"""
    return {str(k): conv_attrib(v) for (k, v) in attribs.items()}


#####################################################################
# Embed Asset Helpers
providers = bootstrap_basic(cache=cache.Cache())


@dc.dataclass(frozen=True)
class Embedded:
    html: HTML
    title: str
    provider: str


def get_embed_url(url: str, width: int = 960, height: int = 540) -> Embedded:
    """Return html for an embeddable URL"""
    try:
        r = providers.request(url, maxwidth=width, maxheight=height)
    except ProviderException:
        # add NoEmbed to the list and try again
        try:
            log.debug("Initializing NoEmbed OEmbed provider")
            bootstrap_noembed(registry=providers)
            r = providers.request(url, maxwidth=width, maxheight=height)
        except ProviderException as e:
            raise ARError(
                f"No embed provider found for URL '{url}' - is there an active internet connection?"
            ) from e

    return Embedded(
        html=r["html"],
        title=r.get("title", "Title"),
        provider=r.get("provider_name", "Embedding"),
    )



"""
Arakawa Processors

API for processing Views, e.g. rendering it locally and publishing to a remote server
"""

from __future__ import annotations

from arakawa.view import Blocks, BlocksT

from .file_store import B64FileEntry
from .processors import (
    ConvertPydantic,
    ExportHTMLInlineAssets,
    ExportHTMLStringInlineAssets,
    ExportHTMLStringInlineNonResizableAssets,
    PreProcessView,
)
from .types import Formatting, Pipeline, ViewState


################################################################################
def save_report(
    blocks: BlocksT,
    path: str,
    open: bool = False,
    name: str | None = None,
    formatting: Formatting | None = None,
    cdn_base: str | None = None,
    standalone: bool = False,
) -> None:
    """Save a report as an HTML file.

    Args:
        blocks: A `Blocks` object or a list of Blocks.
        path: A file path to store the document.
        open: Open in your browser after creating. Default to False.
        name: A name of a report. Optional. Uses path if not provided.
        formatting: Sets the basic app styling.
        cdn_base: Base URL of CDN. Defaults to None.
        standalone: Whether or not to inline assets in an HTML instead of loading via CDN or not. Defaults to False.
    """
    s = ViewState(blocks=Blocks.wrap_blocks(blocks), file_entry_klass=B64FileEntry)

    _ = (
        Pipeline(s)
        .pipe(PreProcessView(is_finalized=True))
        .pipe(ConvertPydantic())
        .pipe(
            ExportHTMLInlineAssets(
                path=path,
                open=open,
                name=name or "Report",
                formatting=formatting,
                cdn_base=cdn_base,
                standalone=standalone,
            )
        )
        .result
    )


def stringify_report(
    blocks: BlocksT,
    name: str | None = None,
    formatting: Formatting | None = None,
    cdn_base: str | None = None,
    resizable: bool = True,
    standalone: bool = False,
) -> str:
    """Stringify a report as an HTML string.

    Args:
        blocks: A `Blocks` object or a list of Blocks.
        name: A name of a report. Optional. Uses path if not provided.
        formatting: Sets the basic app styling.
        cdn_base: Base URL of CDN. Defaults to None.
        resizable: Wether or not to allow make an iframed report resizable or not. Defaults to True.
        standalone: Whether or not to inline assets in an HTML instead of loading via CDN or not. Defaults to False.
    """
    s = ViewState(blocks=Blocks.wrap_blocks(blocks), file_entry_klass=B64FileEntry)
    klass = (
        ExportHTMLStringInlineAssets
        if resizable
        else ExportHTMLStringInlineNonResizableAssets
    )
    export = klass(
        name=name or "Report",
        formatting=formatting,
        cdn_base=cdn_base,
        standalone=standalone,
    )
    return (
        Pipeline(s)
        .pipe(PreProcessView(is_finalized=True))
        .pipe(ConvertPydantic())
        .pipe(export)
        .result
    )



"""Asset-based blocks"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, cast

import pandas as pd
from pandas.io.formats.style import Styler

from arakawa.common.df_processor import to_df
from arakawa.types import NPath

from .base import BlockId, DataBlock

if TYPE_CHECKING:
    from arakawa.processors.file_store import FileEntry


class AssetBlock(DataBlock):
    """
    AssetBlock objects form basis of all File-related blocks (abstract class, not exported)
    """

    _prev_entry: FileEntry | None = None

    # TODO - we may need to support file here as well to handle media, etc.
    def __init__(
        self,
        data: Any = None,
        file: Path | None = None,
        caption: str | None = None,
        name: BlockId | None = None,
        label: str | None = None,
        **kwargs,
    ):
        # storing objects for delayed upload
        super().__init__(name=name, label=label, **kwargs)
        self.data = data
        self.file = file
        self.caption = caption


class Media(AssetBlock):
    """
    The Media block allows you to include images, GIFs, video and audio in your apps. If the file is in a supported format, it will be displayed inline in your app.

    To include an image, you can use `ar.Media` and pass the path.

    !!! note
        Supported video, audio and image formats depend on the browser used to view the report. MP3, MP4, and all common image formats are generally supported by modern browsers
    """

    _tag = "Media"

    def __init__(
        self,
        file: NPath,
        name: BlockId | None = None,
        label: str | None = None,
        caption: str | None = None,
    ):
        """
        Args:
            file: A ath to a file to attach to the report (e.g. a JPEG image)
            name: A unique name for the block to reference when adding text or embedding (optional)
            caption: A caption to display below the file (optional)
            label: A label used when displaying the block (optional)
        """
        file = Path(file).expanduser()
        super().__init__(file=file, name=name, caption=caption, label=label)


class Attachment(AssetBlock):
    """
    If you want to include static files like PDFs or Excel docs in your app, use the `ar.Attachment` block.

    You can also pass in a Python object directly. Once you upload the app, your users will be able to explore and download these attachments.

    !!! tip
        To attach streamable / viewable video, audio or images, use the `ar.Media` block instead
    """

    _tag = "Attachment"

    def __init__(
        self,
        data: Any = None,
        file: NPath | None = None,
        filename: str | None = None,
        caption: str | None = None,
        name: BlockId | None = None,
        label: str | None = None,
    ):
        """
        Args:
            data: A python object to attach to the report (e.g. a dictionary)
            file: A path to a file to attach to the report (e.g. a csv file)
            filename: A name to be used when downloading the file (optional)
            caption: A caption to display below the file (optional)
            name: A unique name for the block to reference when adding text or embedding (optional)
            label: A label used when displaying the block (optional)

        !!! note

            Either `data` or `file` must be provided
        """
        if file:
            file = Path(file).expanduser()
            filename = filename or file.name
        elif data:
            filename = filename or "test.data"

        super().__init__(
            data=data,
            file=cast(Optional[Path], file),
            filename=filename,
            name=name,
            caption=caption,
            label=label,
        )


class Plot(AssetBlock):
    """
    Arakawa supports all major Python visualization libraries, allowing you to add interactive plots and visualizations to your app.

    The `ar.Plot` block takes a plot object from one of the supported Python visualization libraries and renders it in your app.

    !!! info
        Arakawa will automatically wrap your visualization or plot in a `ar.Plot` block if you pass it into your app directly.
    """

    _tag = "Plot"

    def __init__(
        self,
        data: Any,
        caption: str | None = None,
        responsive: bool = True,
        scale: float = 1.0,
        name: BlockId | None = None,
        label: str | None = None,
    ):
        """
        Args:
            data: A `plot` object to attach
            caption: A caption to display below the plot (optional)
            responsive: Whether the plot should automatically be resized to fit, set to False if your plot looks odd (optional, default: True)
            scale: Set the scaling factor for the plt (optional, default = 1.0)
            name: A unique name for the block to reference when adding text or embedding (optional)
            label: A label used when displaying the block (optional)
        """
        super().__init__(
            data=data,
            caption=caption,
            responsive=responsive,
            scale=scale,
            name=name,
            label=label,
        )


class Table(AssetBlock):
    """
    Table blocks store the contents of a DataFrame as a HTML `table` whose style can be customized using pandas' `Styler` API.

    !!! tip
        `Table` is the best option for displaying multidimensional DataFrames, as `DataTable` will flatten your data.
    """

    # NOTE - Tables are stored as HTML fragment files rather than inline within the Report document

    _tag = "Table"

    def __init__(
        self,
        data: pd.DataFrame | Styler,
        caption: str | None = None,
        name: BlockId | None = None,
        label: str | None = None,
    ):
        """
        Args:
            data: A pandas `Styler` instance or dataframe to generate the table from
            caption: A caption to display below the table (optional)
            name: A unique name for the block to reference when adding text or embedding (optional)
            label: A label used when displaying the block (optional)
        """
        super().__init__(data=data, caption=caption, name=name, label=label)


class DataTable(AssetBlock):
    """
    The DataTable block takes a pandas DataFrame and renders an interactive, sortable, searchable table in your app, along with advanced analysis options such as exploring data through [SandDance](https://www.microsoft.com/en-us/research/project/sanddance/).

    It supports large datasets and viewers can also download the table from the website as a CSV or Excel file.

    !!! tip
        `Table` is the best option for displaying multidimensional DataFrames, as `DataTable` will flatten your data.
    """

    _tag = "DataTable"

    def __init__(
        self,
        df: pd.DataFrame,
        caption: str | None = None,
        name: BlockId | None = None,
        label: str | None = None,
    ):
        """
        Args:
            df: A pandas dataframe to attach to the report
            caption: A caption to display below the plot (optional)
            name: A unique name for the block to reference when adding text or embedding (optional)
            label: A label used when displaying the block (optional)
        """
        # create a copy of the df to process
        df = to_df(df)
        super().__init__(data=df, caption=caption, name=name, label=label)
        # TODO - support pyarrow schema for local reports
        (rows, columns) = df.shape
        self._add_attributes(rows=rows, columns=columns)



import string
import sys
from typing import Any

from loguru._logger import Core as _Core
from loguru._logger import Logger as _Logger


def get_logger():
    logger = _Logger(
        core=_Core(),
        exception=None,
        depth=0,
        record=False,
        lazy=False,
        colors=False,
        raw=False,
        capture=True,
        patchers=[],
        extra={},
    )
    logger.add(sys.stderr, level="WARNING")
    return logger


log = get_logger()

_have_setup_logging: bool = False


def enable_logging():
    """Enable logging for debug purposes"""
    global _have_setup_logging

    log.remove(0)
    log.add(sys.stderr, level="DEBUG")

    _have_setup_logging = True


def print_debug_info():
    """Print useful debugging information"""
    fields = {}

    # Known dependencies
    import numpy as np
    import pandas as pd
    import pyarrow as pa

    fields["pandas_version"] = pd.__version__
    fields["numpy_version"] = np.__version__
    fields["pyarrow_version"] = pa.__version__

    for _k, _v in fields.items():
        pass


################################################################################
# Output
def open_in_browser(url: str):
    """Open the given URL in the user's browser"""
    from arakawa.ipython.environment import get_environment

    environment = get_environment()

    # TODO - this is a bit of a hack, but works for now. JupyterLab (Codespaces) doesn't support webbrowser.open.
    if (
        environment.is_notebook_environment
        and not environment.can_open_links_from_python
    ):
        ip = environment.get_ipython()
        ip.run_cell_magic("javascript", "", f'window.open("{url}", "_blank")')
    else:
        import webbrowser

        webbrowser.open(url, new=1)


class MarkdownFormatter(string.Formatter):
    """Support {:l} and {:cmd} format fields"""

    in_jupyter: bool

    def __init__(self, in_jupyter: bool):
        self.in_jupyter = in_jupyter
        super().__init__()

    def format_field(self, value: Any, format_spec: str) -> Any:
        if format_spec.endswith("l"):
            if self.in_jupyter:
                value = f"<a href='{value}' target='_blank'>here</a>"
            else:
                value = f"at {value}"
            format_spec = format_spec[:-1]
        elif format_spec.endswith("cmd"):
            value = f"!{value}" if self.in_jupyter else value
            format_spec = format_spec[:-3]
        return super().format_field(value, format_spec)


def display_msg(text: str, **params: str):
    from arakawa.ipython.environment import get_environment

    environment = get_environment()

    msg = MarkdownFormatter(environment.is_notebook_environment).format(text, **params)
    if environment.is_notebook_environment:
        from IPython.display import Markdown, display

        display(Markdown(msg))
    else:
        print(msg)  # noqa: T201
